/** @example  ex/ios/EventCallbackProtocol.h
 */

/*
 *  Copyright 2014-2025 Solace Corporation. All rights reserved.
 */

#import <Foundation/Foundation.h>
#import "solclient/solCache.h"
#import "solclient/solClient.h"

/**
 *  The EventCallbackProtocol protocol defines the methods used for various
 *  event callbacks by the examples in the application. Each method has a C
 *  function counterpart, as the CCSMP API does not support passing Objective C
 *  methods as function pointers.
 *
 *  Each example that uses an event callback must implement this protocol and
 *  pass the corresponding C function for that callback to the API. The user
 *  data for the callback must be an instance of the callbackBridge_t struct,
 *  which must contain a reference to the Objective C object that corresponds to
 *  the example as well as any additional user data. The C function will then
 *  call the appropriate event callback method in that Objective C object.
 *
 *  Note that because the main Example class implements this protocol, and all
 *  examples inherit from the main Example class, it is only necessary for
 *  particular examples to re-implement the methods defined by this protocol
 *  when the default behaviour of the callbacks must be changed.
 */
@protocol EventCallbackProtocol <NSObject>

/**
 *  Callback for session events. The callback is registered for a session and is
 *  called whenever a session event occurs.
 *
 *  @param opaqueSession_p A pointer to the session to which the event applies.
 *                         This pointer is never NULL.
 *  @param eventInfo_p     A pointer to information about the event, such as the
 *                         event type. This pointer is never NULL.
 *  @param user_p          A pointer to opaque user data.
 */
- (void)eventCallbackWithSession:(solClient_opaqueSession_pt)opaqueSession_p
                       eventInfo:
                           (solClient_session_eventCallbackInfo_pt)eventInfo_p
                        userData:(void *)user_p;

/**
 *  The corresponding C function for
 *  eventCallbackWithSession:eventInfo:userData. See the documentation for that
 *  method for more information.
 *
 *  @param opaqueSession_p A pointer to the session to which the event applies.
 *                         This pointer is never NULL.
 *  @param eventInfo_p     A pointer to information about the event, such as the
 *                         event type. This pointer is never NULL.
 *  @param user_p          A pointer to an instance of the callbackBridge_t
 *                         struct.
 */
void eventCallback(solClient_opaqueSession_pt opaqueSession_p,
                   solClient_session_eventCallbackInfo_pt eventInfo_p,
                   void *user_p);
/**
 *  Callback for cache events. The callback is provided when making a
 *  non-blocking cache request to perform actions when a cache event occurs.
 *
 *  @param opaqueSession_p A pointer to the session to which the event applies.
 *                         This pointer is never NULL.
 *  @param eventInfo_p     A pointer to information about the event, such as the
 *                         event type. This pointer is never NULL.
 *  @param user_p          A pointer to opaque user data.
 */
- (void)cacheEventCallbackWithSession:
            (solClient_opaqueSession_pt)opaqueSession_p
                            eventInfo:(solCache_eventCallbackInfo_pt)eventInfo_p
                             userData:(void *)user_p;

/**
 *  The corresponding C function for
 *  cacheEventCallbackWithSession:eventInfo:userData. See the documentation for
 *  that method for more information.
 *
 *  @param opaqueSession_p A pointer to the session to which the event applies.
 *                         This pointer is never NULL.
 *  @param eventInfo_p     A pointer to information about the event, such as the
 *                         event type. This pointer is never NULL.
 *  @param user_p          A pointer to an instance of the callbackBridge_t
 *                         struct.
 */
void cacheEventCallback(solClient_opaqueSession_pt opaqueSession_p,
                        solCache_eventCallbackInfo_pt eventInfo_p,
                        void *user_p);

/**
 *  Callback for session events. The callback is registered for a session and is
 *  called whenever a session event occurs. This performance event callback does
 *  nothing.
 *
 *  @param opaqueSession_p A pointer to the session to which the event applies.
 *                         This pointer is never NULL.
 *  @param eventInfo_p     A pointer to information about the event, such as the
 *                         event type. This pointer is never NULL.
 *  @param user_p          A pointer to opaque user data.
 */
- (void)eventPerfCallbackWithSession:(solClient_opaqueSession_pt)opaqueSession_p
                           eventInfo:(solClient_session_eventCallbackInfo_pt)
                           eventInfo_p
                            userData:(void *)user_p;

/**
 *  The corresponding C function for
 *  eventPerfCallbackWithSession:eventInfo:userData. See the documentation for
 *  that method for more information.
 *
 *  @param opaqueSession_p A pointer to the session to which the event applies.
 *                         This pointer is never NULL.
 *  @param eventInfo_p     A pointer to information about the event, such as the
 *                         event type. This pointer is never NULL.
 *  @param user_p          A pointer to an instance of the callbackBridge_t
 *                         struct.
 */
void eventPerfCallback(solClient_opaqueSession_pt opaqueSession_p,
                       solClient_session_eventCallbackInfo_pt eventInfo_p,
                       void *user_p);

@end
