/** @example ex/ios/Example.h
 */

/*
 *  Copyright 2014-2025 Solace Corporation. All rights reserved.
 */

#import "EventCallbackProtocol.h"
#import "ExampleInterface.h"
#import "MessageReceiveCallbackProtocol.h"
#import "LogCallbackProtocol.h"
#import <mach/clock.h>
#import <mach/mach.h>
#import "ParameterInterface.h"
#import <pthread.h>
#import "solclient/solCache.h"
#import "solclient/solClient.h"
#import "solclient/solClientMsg.h"

/**
 *  The Example class is the superclass for all examples in the application.
 */
@interface Example : NSObject <MessageReceiveCallbackProtocol,
                               EventCallbackProtocol, LogCallbackProtocol>

/**
 *  Common values that can be used by examples.
 */

/**
 *  A sample topic.
 */
#define COMMON_MY_SAMPLE_TOPIC "my/sample/topic"

/**
 *  Sample format for SEMP topics.
 */
#define COMMON_SEMP_TOPIC_FORMAT "#SEMP/%s/SHOW"

/**
 *  Sample attachment text.
 */
#define COMMON_ATTACHMENT_TEXT "my attached data"

// Thread definitions and typedefs
#define DEFAULT_THREAD_RETURN_ARG (NULL)
#define _NULL_THREAD_ID (0)
typedef void *(*FP)(void *);
typedef void *threadRetType;
typedef threadRetType (*threadFn_pt)(void *param_p);

/**
 *  @struct callbackBridge
 *  Allows C callback functions to reference Objective C methods.
 */
typedef struct callbackBridge {
    /**
     *  The example object containing the Objective C callback method.
     */
    void *example_p;

    /**
     *  Any user data passed to the callback.
     */
    void *user_p;
} callbackBridge_t;

/**
 * @struct contextThreadInfo
 * Holds context thread information.
 */
typedef struct contextThreadInfo {

    /**
     *  Session context.
     */
    solClient_opaqueContext_pt context_p;

    /**
     *  Indicates whether the thread has been started.
     */
    volatile int contextThreadStarted;

    /**
     *  Indicates whether the thread should be stopped.
     */
    volatile int stopContextThread;

    /**
     *  Return code from solClient_context_processEvents.
     */
    int rc;

    /**
     *  Thread handle.
     */
    pthread_t handle;
} contextThreadInfo_t;

/**
 * @struct ThreadInfo
 * Holds thread information.
 */
typedef struct threadInfo {

    /**
     *  Indicates whether the thread has been started.
     */
    volatile int threadStarted; //*< 1 when the thread has been started

    /**
     *  Indicates whether the thread should be stopped.
     */
    volatile int stopThread; //*< set to 1 to stop the context thread

    /**
     *  Return code from solClient_context_processEvents.
     */
    int rc;

    /**
     *  Thread handle.
     */
    pthread_t handle; //*< thread handle

    /**
     *  User context.
     */
    void *user_p;
} threadInfo_t;

/**
 *  The name of the example.
 */
@property NSString *name;

/**
 *  The description of the example.
 */
@property (copy, atomic)NSString *description;

/**
 *  The configuration parameters associated with the example.
 */
@property ParameterInterface *parameters;

/**
 *  The example interface associated with the example.
 */
@property ExampleInterface *exampleInterface;

/**
 *  Indicates whether the example is currently running.
 */
@property BOOL running;

/**
 *  Indicates whether the UI has requested that the example be started.
 */
@property BOOL requestRun;

/**
 *  Indicates whether the UI has requested that the example be started.
 */
@property BOOL requestCancel;

/**
 *  The default callback bridge used by most callbacks. The user data pointer is
 *  always null.
 */
@property callbackBridge_t *nullBridge_p;

/**
 *  Initializes an instance of the Example class with an example interface.
 *
 *  @param exampleInterface The ExampleInterface object with which to initialize
 *                          the Example object.
 *
 *  @return An Example object.
 */
- (id)initWithExampleInterface:(ExampleInterface *)exampleInterface;

/**
 *  Starts the example.
 */
- (void)start;

/**
 *  This method is called when the example starts. It contains the actual
 *  example code, and must be overridden by subclasses. However, the run method
 *  for each subclass must still call [super run].
 */
- (void)run;

/**
 *  Cancels the example.
 */
- (void)cancel;

/**
 *  This method is called when the example ends. It contains the example
 *  cleanup code, and must be called at the end of the run method.
 */
- (void)cleanup;

/**
 *  Prints a string to the text view of the currently running example.
 *
 *  @param str The string to print.
 */
- (void)printToTextView:(NSString *)str;

/**
 *  Logs the CCSMP API version.
 */
- (void)logCCSMPVersion;

/**
 *  Sets the logging level of the CCSMP API based on the value of the logging
 *  parameters.
 */
- (void)setLoggingLevel;

/**
 *  Logs an error message based on a specified return code and error string.
 *
 *  @param rc       The return code to be logged.
 *  @param errorStr The error message to be logged.
 */
- (void)handleErrorWithReturnCode:(solClient_returnCode_t)rc
                      errorString:(const char *)errorStr;

/**
 *  Creates a session and connects the session. The default callback bridge is
 *  used as user data.
 *
 *  @param context_p       A pointer to the context in which the session is to
 *                         be created.
 *  @param session_p       A pointer to a session pointer. This session pointer
 *                         points to the newly created session upon completion
 *                         of this function.
 *  @param msgCallback_p   The message callback function to use when creating
 *                         the session.
 *  @param eventCallback_p The event callback function to use when creating the
 *                         session.
 *
 *  @return The return code associated with session creation and connection.
 */
- (solClient_returnCode_t)
    createAndConnectSessionWithContext:(solClient_opaqueContext_pt)context_p
                               session:(solClient_opaqueSession_pt *)session_p
                       messageCallback:
                           (solClient_session_rxMsgCallbackFunc_t)msgCallback_p
                         eventCallback:(solClient_session_eventCallbackFunc_t)
    eventCallback_p;

/**
 *  Creates a session and connects the session.
 *
 *  @param context_p       A pointer to the context in which the session is to
 *                         be created.
 *  @param session_p       A pointer to a session pointer. This session pointer
 *                         points to the newly created session upon completion
 *                         of this function.
 *  @param msgCallback_p   The message callback function to use when creating
 *                         the session.
 *  @param eventCallback_p The event callback function to use when creating the
 *                         session.
 *  @param user_p          A pointer to an instance of the callbackBridge_t
 *                         struct.
 *
 *  @return The return code associated with session creation and connection.
 */
- (solClient_returnCode_t)
    createAndConnectSessionWithContext:(solClient_opaqueContext_pt)context_p
                               session:(solClient_opaqueSession_pt *)session_p
                       messageCallback:
                           (solClient_session_rxMsgCallbackFunc_t)msgCallback_p
                         eventCallback:(solClient_session_eventCallbackFunc_t)
                         eventCallback_p
                              userData:(callbackBridge_t *)user_p;

/**
 *  Publishes a message to the specified topic using the specified session.
 *
 *  @param session_p    A pointer to the session.
 *  @param topic_p      The topic to publish on.
 *  @param deliveryMode The delivery mode used to publish the message.
 *
 *  @return The return code associated with the message publication.
 */
- (solClient_returnCode_t)
    publishMessageWithSession:(solClient_opaqueSession_pt)session_p
                        topic:(char *)topic_p
                 deliveryMode:(solClient_uint32_t)deliveryMode;

/**
 *  Starts a new thread.
 *
 *  @param fp  The thread function to be executed.
 *  @param arg The argument to pass to the thread function.
 *
 *  @return The thread handle.
 */
- (pthread_t)startThreadWithFP:(FP)fp argument:(void *)arg;

/**
 *  Waits on the thread.
 *
 *  @param handle Thread handle to wait on.
 */
- (void)waitOnThread:(pthread_t)handle;

/**
 *  Function that runs in the context thread.
 *
 *  @param contextInfo_p A pointer to context thread information.
 *
 *  @return Thread return code.
 */
- (threadRetType)contextThread:(void *)contextInfo_p;

/**
 *  The corresponding C function for contextThread:. See the documentation for
 *  that method for more information.
 *
 *  @param user_p Context information.
 *
 *  @return Thread return code code.
 */
threadRetType contextThread(void *user_p);

/**
 *  Starts the context thread.
 *
 *  @param info_p Context thread information.
 *
 *  @return Return code.
 */
- (int)startContextThread:(contextThreadInfo_t *)info_p;

/**
 *  Stops the context thread.
 *
 *  @param info_p Context information.
 */
- (void)stopContextThread:(contextThreadInfo_t *)info_p;

/**
 *  Gets the current time in microseconds.
 *
 *  @return The current time in microseconds.
 */
- (UInt64)getTimeInUs;

/**
 *  Gets the usage time.
 *
 *  @param userTime_p   User time pointer.
 *  @param systemTime_p System time pointer.
 */
- (void)getUsageTimeWithUserTime:(long long *)userTime_p
                      systemTime:(long long *)systemTime_p;

@end
