/** @example ex/ios/MessageReceiveCallbackProtocol.h
 */

/*
 *  Copyright 2014-2025 Solace Corporation. All rights reserved.
 */

#import <Foundation/Foundation.h>
#import "solclient/solClient.h"

/**
 *  The MessageReceiveCallbackProtocol.h protocol defines the methods used for
 *  various message callbacks by the examples in the application. Each method
 *  has a C function counterpart, as the CCSMP API does not support passing
 *  Objective C methods as function pointers.
 *
 *  This protocol is used in the same manner as the EventCallbackProtocol.
 *  Please see the documentation for that protocol for more information.
 */
@protocol MessageReceiveCallbackProtocol <NSObject>

/**
 *  Callback for received messages by a session. The callback is registered
 *  for a session and is called whenever a message is received.
 *
 *  @param opaqueSession_p A pointer to the session receiving the message.
 *                         This pointer is never NULL.
 *  @param msg_p           A pointer to the received message. This pointer is
 *                         never NULL.
 *  @param user_p          A pointer to opaque user data.
 */
- (solClient_rxMsgCallback_returnCode_t)
    messageReceiveCallbackWithSession:
        (solClient_opaqueSession_pt)opaqueSession_p
                              message:(solClient_opaqueMsg_pt)msg_p
                             userData:(void *)user_p;

/**
 *  The corresponding C function for
 *  messageReceiveCallbackWithSession:message:userData. See the documentation
 *  for that method for more information.
 *
 *  @param opaqueSession_p A pointer to the session receiving the message.
 *                         This pointer is never NULL.
 *  @param msg_p           A pointer to the received message. This pointer is
 *                         never NULL.
 *  @param user_p          A pointer to an instance of the callbackBridge_t
 *                         struct.
 */
solClient_rxMsgCallback_returnCode_t
messageReceiveCallback(solClient_opaqueSession_pt opaqueSession_p,
                       solClient_opaqueMsg_pt msg_p, void *user_p);

/**
 *  Callback for received messages by a session. The callback is registered
 *  for a session and is called whenever a message is received. This callback
 *  logs the contents of the message. The user_p parameter is assumed to point
 *  to a NULL terminated char array containing a user specified name for the
 *  Session.
 *
 *  @param opaqueSession_p A pointer to the session receiving the message.
 *                         This pointer is never NULL.
 *  @param msg_p           A pointer to the received message. This pointer is
 *                         never NULL.
 *  @param user_p          A pointer to opaque user data.
 */
- (solClient_rxMsgCallback_returnCode_t)
    messageReceivePrintMsgCallbackWithSession:
        (solClient_opaqueSession_pt)opaqueSession_p
                                      message:(solClient_opaqueMsg_pt)msg_p
                                     userData:(void *)user_p;

/**
 *  The corresponding C function for
 *  messageReceivePrintMsgCallbackWithSession:message:userData. See the
 *  documentation for that method for more information.
 *
 *  @param opaqueSession_p A pointer to the session receiving the message.
 *                         This pointer is never NULL.
 *  @param msg_p           A pointer to the received message. This pointer is
 *                         never NULL.
 *  @param user_p          A pointer to an instance of the callbackBridge_t
 *                         struct.
 */
solClient_rxMsgCallback_returnCode_t
messageReceivePrintMsgCallback(solClient_opaqueSession_pt opaqueSession_p,
                               solClient_opaqueMsg_pt msg_p, void *user_p);

/**
 *  Callback for received messages by a session. The callback is registered
 *  for a session and is called whenever a message is received. This performance
 *  callback does nothing.
 *
 *  @param opaqueSession_p A pointer to the session receiving the message.
 *                         This pointer is never NULL.
 *  @param msg_p           A pointer to the received message. This pointer is
 *                         never NULL.
 *  @param user_p          A pointer to opaque user data.
 */
- (solClient_rxMsgCallback_returnCode_t)
    messageReceivePerfCallbackWithSession:
        (solClient_opaqueSession_pt)opaqueSession_p
                                  message:(solClient_opaqueMsg_pt)msg_p
                                 userData:(void *)user_p;

/**
 *  The corresponding C function for messageReceivePerfCallback:msg_p:userData.
 *  See the documentation for that method for more information.
 *
 *  @param opaqueSession_p A pointer to the session receiving the message.
 *                         This pointer is never NULL.
 *  @param msg_p           A pointer to the received message. This pointer is
 *                         never NULL.
 *  @param user_p          A pointer to an instance of the callbackBridge_t
 *                         struct.
 */
solClient_rxMsgCallback_returnCode_t
messageReceivePerfCallback(solClient_opaqueSession_pt opaqueSession_p,
                           solClient_opaqueMsg_pt msg_p, void *user_p);

@end
