/**
 *  ParameterViewController.m
 *
 *  Copyright 2014-2025 Solace Corporation. All rights reserved.
 */

#import "LogViewController.h"
#import "Example.h"
#import "ParameterViewController.h"

@implementation ParameterViewController

@synthesize example;
@synthesize parameterTableView;

- (void)prepareForSegue:(UIStoryboardSegue *)segue sender:(id)sender {
    // Pass example to the next view controller
    LogViewController *newController =
        (LogViewController *)segue.destinationViewController;
    newController.example = example;
}

- (IBAction)paramEditEnd:(id)sender {
    // Retrieve the tag of the cell that contains the sender (a text field).
    // Substract 1 because the index of the corresponding parameter is 0-based.
    int index = (int)[[sender superview] tag] - 1;

    // Save the value in the parameter object
    [example.parameters parameterAtIndex:index].value = [sender text];
    [sender resignFirstResponder];
}

- (BOOL)textFieldShouldReturn:(UITextField *)textField {
    // Ensure that keyboard disappears when the enter button is pressed in a
    // text field
    [textField resignFirstResponder];
    return YES;
}

- (IBAction)switchValueChanged:(id)sender {
    // Retrieve the tag of the cell that contains the sender (a textfield).
    // Substract 1 because the index of the corresponding parameter is 0-based.
    int index = (int)[[sender superview] tag] - 1;
    UISwitch *switchObject = sender;
    // Save the value in the parameter object
    [example.parameters parameterAtIndex:index].value =
        switchObject.on ? @"1" : @"0";
    [sender resignFirstResponder];
}

- (UITableViewCell *)tableView:(UITableView *)tableView
         cellForRowAtIndexPath:(NSIndexPath *)indexPath {
    // Create new cells as needed for a given entry in the table view
    int index = (int)indexPath.row;
    UITableViewCell *cell;
    // Create extra cells for padding if index exceeds number of parameters
    if (index >= [self.example.parameters numOfParameters]) {
        cell = [tableView dequeueReusableCellWithIdentifier:@"emptyCell"];
        if (cell == nil) {
            cell = [[UITableViewCell alloc]
                  initWithStyle:UITableViewCellStyleDefault
                reuseIdentifier:@"emptyCell"];
        }
    } else {
        // Otherwise, create the appropriate cell for the parameter type
        Parameter *parameter = [example.parameters parameterAtIndex:index];
        switch ([parameter type]) {
        default:
        case StringParameter:
            cell = [self createStringCell:tableView parameterToAdd:parameter];
            break;
        case BooleanParameter:
            cell = [self createBooleanCell:tableView parameterToAdd:parameter];
            break;
        case PickerParameter:
            cell = [self createPickerCell:tableView parameterToAdd:parameter];
            break;
        }
        // Hide the cell's textLabel or else our custom control in the cell
        // won't be visible
        cell.textLabel.hidden = true;
    }

    // Tag the cell with the index of the parameter so we know what cell links
    // to which parameter
    // Since tag cannot be null, we'll use a 1 based index for tags.
    cell.contentView.tag = index + 1;

    return cell;
}

- (NSInteger)tableView:(UITableView *)tableView
    numberOfRowsInSection:(NSInteger)section {
    // The number of rows in the table view corresponds to the number of
    // examples plus a number of padding cells
    return [self.example.parameters numOfParameters] + PADDING_CELLS;
}

- (NSInteger)numberOfSectionsInTableView:(UITableView *)tableView {
    // The table view only has one section
    return 1;
}

- (UITableViewCell *)createStringCell:(UITableView *)tableView
                       parameterToAdd:(Parameter *)parameter {
    // Use existing, unused cell if possible
    UITableViewCell *cell =
        [tableView dequeueReusableCellWithIdentifier:@"stringCell"];
    if (cell == nil) {
        cell =
            [[UITableViewCell alloc] initWithStyle:UITableViewCellStyleDefault
                                   reuseIdentifier:@"stringCell"];
    }

    // Set label and default value of text field
    UILabel *paramNameLabel = (UILabel *)cell.contentView.subviews[1];
    [paramNameLabel setText:[parameter displayName]];
    UITextField *textField = (UITextField *)cell.contentView.subviews[0];
    [textField setText:[parameter value]];

    return cell;
}

- (UITableViewCell *)createBooleanCell:(UITableView *)tableView
                        parameterToAdd:(Parameter *)parameter {
    // Use existing, unused cell if possible
    UITableViewCell *cell =
        [tableView dequeueReusableCellWithIdentifier:@"booleanCell"];
    if (cell == nil) {
        cell =
            [[UITableViewCell alloc] initWithStyle:UITableViewCellStyleDefault
                                   reuseIdentifier:@"booleanCell"];
    }

    // Set label and default value of switch
    UILabel *paramNameLabel = (UILabel *)cell.contentView.subviews[0];
    paramNameLabel.text = [parameter displayName];
    UISwitch *booleanSwitch = (UISwitch *)cell.contentView.subviews[1];
    [booleanSwitch setOn:[[parameter value] isEqualToString:@"1"]];

    return cell;
}

- (UITableViewCell *)createPickerCell:(UITableView *)tableView
                       parameterToAdd:(Parameter *)parameter {
    // Use existing, unused cell if possible
    UITableViewCell *cell =
        [tableView dequeueReusableCellWithIdentifier:@"pickerCell"];
    if (cell == nil) {
        cell =
            [[UITableViewCell alloc] initWithStyle:UITableViewCellStyleDefault
                                   reuseIdentifier:@"pickerCell"];
    }

    // Set label, possible values and default value of picker
    UILabel *paramNameLabel = (UILabel *)cell.contentView.subviews[0];
    paramNameLabel.text = [parameter displayName];
    UIPickerView *picker = (UIPickerView *)cell.contentView.subviews[1];
    picker.dataSource = parameter;
    picker.delegate = parameter;
    [picker selectRow:parameter.index inComponent:0 animated:YES];

    return cell;
}

@end
