/** @example ex/ios/examples/EventMonitorExample.m
 */

/*
 *  Copyright 2009-2025 Solace Corporation. All rights reserved.
 */

#import "EventMonitorExample.h"

@implementation EventMonitorExample

- (id)initWithExampleInterface:(ExampleInterface *)exampleInterface {
    self = [super initWithExampleInterface:exampleInterface];

    // Set example name and description
    self.name = @"EventMonitor";
    self.description = @"Demonstrates how to monitor appliance events using an "
        @"event subscription";

    // Setup example parameters
    [self.parameters addParameter:PARAMETER_NUM_MESSAGES];

    return self;
}

- (solClient_rxMsgCallback_returnCode_t)
    messageReceiveCallbackWithSession:
        (solClient_opaqueSession_pt)opaqueSession_p
                              message:(solClient_opaqueMsg_pt)msg_p
                             userData:(void *)user_p {
    solClient_returnCode_t rc = SOLCLIENT_OK;
    solClient_destination_t destination;

    void *binaryAttachmentBuffer_p;
    solClient_uint32_t binaryAttachmentBufferSize;

    if ((rc = solClient_msg_getDestination(
             msg_p, &destination, sizeof(destination))) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_msg_getDestination()"];
        return SOLCLIENT_CALLBACK_OK;
    }

    if ((rc = solClient_msg_getBinaryAttachmentPtr(
             msg_p, &binaryAttachmentBuffer_p, &binaryAttachmentBufferSize)) !=
        SOLCLIENT_OK) {
        [self
            handleErrorWithReturnCode:rc
                          errorString:"solClient_msg_getBinaryAttachmentPtr()"];
        return SOLCLIENT_CALLBACK_OK;
    }

    // Log event message
    solClient_log(SOLCLIENT_LOG_NOTICE, "*** Event Message Received ***\n");
    solClient_log(SOLCLIENT_LOG_NOTICE, "Topic:\t%s\n", destination.dest);
    solClient_log(SOLCLIENT_LOG_NOTICE, "Event:\t%s\n",
                  (char *)binaryAttachmentBuffer_p);

    // Returning SOLCLIENT_CALLBACK_OK causes the API to free the memory
    // used by the message. This is important to avoid leaks.
    return SOLCLIENT_CALLBACK_OK;
}

- (solClient_returnCode_t)triggerSecondaryConnectionWithContext:
                              (solClient_opaqueContext_pt)context_p {
    solClient_opaqueSession_pt session_p;
    solClient_returnCode_t rc = SOLCLIENT_OK;

    if ((rc = [self
             createAndConnectSessionWithContext:context_p
                                        session:&session_p
                                messageCallback:messageReceivePrintMsgCallback
                                  eventCallback:eventCallback]) !=
        SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:
                  rc errorString:"createAndConnectSessionWithContext:session:"
                                 "messageCallback:eventCallback:userData:"];
        return rc;
    }

    if ((rc = solClient_session_topicSubscribeExt(
             session_p, SOLCLIENT_SUBSCRIBE_FLAGS_WAITFORCONFIRM,
             COMMON_MY_SAMPLE_TOPIC)) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_session_topicSubscribe()"];
    }

    if ((rc = solClient_session_disconnect(session_p)) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_session_disconnect()"];
    }

    return rc;
}

- (void)run {
    [super run];
    /**
     *  Initialization
     */

    solClient_returnCode_t rc = SOLCLIENT_OK;

    // Context variables
    solClient_opaqueContext_pt context_p;
    solClient_context_createFuncInfo_t contextFuncInfo =
        SOLCLIENT_CONTEXT_CREATEFUNC_INITIALIZER;

    // Session variables
    solClient_opaqueSession_pt session_p;

    // When event message publishing is enabled for a message VPN, clients can
    // subscribe to a specific topic to receive the event messages.
    //
    // Format for Event topics:
    //     #LOG/<level>/<type>/<appliance hostname>/<event name>/...
    char *eventTopicFormat = "#LOG/INFO/CLIENT/%s/CLIENT_CLIENT_CONNECT/>";

    char eventTopic[SOLCLIENT_BUFINFO_MAX_TOPIC_SIZE + 1];
    solClient_field_t routerName;

    // Initialize the API; this must be done prior to first usage
    if ((rc = solClient_initialize(SOLCLIENT_LOG_DEFAULT_FILTER, NULL)) !=
        SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_initialize()"];
        goto notInitialized;
    }

    // Set up logging level and log example and API information
    [self setLoggingLevel];
    solClient_log(
        SOLCLIENT_LOG_NOTICE,
        "EventMonitorExample.m (Copyright 2009-2025 Solace Corporation. All rights reserved.)\n");
    [self logCCSMPVersion];

    // Create a context, and specify that the context thread be created
    // automatically instead of having the application create its own
    // context thread.
    solClient_log(SOLCLIENT_LOG_INFO, "Creating solClient context");
    if ((rc = solClient_context_create(
             SOLCLIENT_CONTEXT_PROPS_DEFAULT_WITH_CREATE_THREAD, &context_p,
             &contextFuncInfo, sizeof(contextFuncInfo))) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_context_create()"];
        goto cleanup;
    }

    /**
     *  Session creation and connection
     */
    solClient_log(SOLCLIENT_LOG_INFO, "Creating solClient sessions.");

    /**
     *  createAndConnectSessionWithContext:session:messageCallback:
     *  eventCallback:userData is a common function used in these samples.
     *  It is a wrapper for solClient_session_create() that applies some
     *  common properties to the Session, some of which are based on the
     *  command options. The wrapper also connects the Session.
     */
    if ((rc = [self createAndConnectSessionWithContext:context_p
                                               session:&session_p
                                       messageCallback:messageReceiveCallback
                                         eventCallback:eventCallback]) !=
        SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:
                  rc errorString:"createAndConnectSessionWithContext:session:"
                                 "messageCallback:eventCallback:userData:"];
        goto cleanup;
    }

    /**
     *  Subscribe to topic
     */

    // Get the hostname of the appliance. The hostname is part of the Topic
    // used to subscribe to Event Messages.
    if ((rc = solClient_session_getCapability(
             session_p, SOLCLIENT_SESSION_PEER_ROUTER_NAME, &routerName,
             sizeof(routerName))) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_session_getCapability()"];
        goto sessionConnected;
    }

    snprintf(eventTopic, sizeof(eventTopic), eventTopicFormat,
             routerName.value.string);

    if ((rc = solClient_session_topicSubscribeExt(
             session_p, SOLCLIENT_SUBSCRIBE_FLAGS_WAITFORCONFIRM,
             eventTopic)) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_session_topicSubscribe()"];
        goto sessionConnected;
    }

    if ((rc = [self triggerSecondaryConnectionWithContext:context_p]) !=
        SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"triggerSecondaryConnection()"];
        goto sessionConnected;
    }

    // Sleep to allow reception of event.
    [NSThread sleepForTimeInterval:1];
    solClient_log(SOLCLIENT_LOG_NOTICE, "Cleaning up.\n");

/**
 *  Cleanup
 */

sessionConnected:
    // Disconnect the session
    if ((rc = solClient_session_disconnect(session_p)) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_session_disconnect()"];
    }

cleanup:
    // Cleanup solclient
    if ((rc = solClient_cleanup()) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc errorString:"solClient_cleanup()"];
    }

notInitialized:
    [self cleanup];
}

@end
