/** @example ex/ios/examples/PerfTestExample.h
 */

/*
 *  Copyright 2007-2025 Solace Corporation. All rights reserved.
 */

#import "Example.h"

/**
 *  The PerfTestExample is capable of performing the following performance
 *  tests:
 *    1. Publisher throughput - will send a number of messages at a given rate
 *       to a Topic.
 *       Program arguments of interest:
 *           NUM_MSGS - specify the number of messages to publish
 *           MSG_RATE - publish message rate (per second)
 *           MSG_SIZE - size of binary payload portion of message
 *           PUB_SUB_MODE - must be 'p'
 *           MULTI_SEND_MODE - "true" to send multiple messages per send call.
 *       (see pubThread: and pubThreadSendMultiple:).
 *    2. Subscriber throughput - will subscribe to messages on a Topic
 *       Program arguments of interest:
 *           NUM_MSGS - specify the number of messages expected to receive. The
 *           program will exit after receiving this number of messages or
 *           more, or if the user presses the Cancel button, whichever comes
 *           first.
 *           PUB_SUB_MODE - must be 's'
 *       (see msgRxCallbackWithSession:bufferInfo:callbackInfo:userData:).
 *    3. Publisher/subscriber throughput - will set up a publisher and
 *       subscriber to a Topic.
 *       Program arguments of interest:
 *           NUM_MSGS - specify the number of messages to publish and expected
 *           to receive.
 *           MSG_RATE - publish message rate (per second)
 *           MSG_SIZE - size of binary payload portion of message
 *           PUB_SUB_MODE - must be 'b'
 *           MULTI_SEND_MODE - "true" to send multiple messages per send call.
 *       (see pubThread:, pubThreadSendMultiple:, and
 *       msgRxCallbackWithSession:bufferInfo:callbackInfo:userData:).
 *
 *  At the end of each test, performance statistics are printed (see
 *  printStatsForSession:).
 */
@interface PerfTestExample : Example

#ifndef DOXYGEN_SHOULD_SKIP_THIS
#define GROUP_SIZE 10
#endif

/**
 *  Indicates whether the example is done receiving messages.
 */
@property int rxDone_s;

/**
 *  Indicates whether the example has timed out.
 */
@property int rxTimeout_s;

/**
 *  Indicates whether the publisher option is enabled.
 */
@property int usePub_s;

/**
 *  Indicates whether the subscriber option is enabled.
 */
@property int useSub_s;

/**
 *  Indicates whether the multi-send option is enabled.
 */
@property int multiSend_s;

/**
 *  Binary payload size for each message.
 */
@property int binaryPayloadSize_s;

/**
 *  Number of messages received.
 */
@property unsigned int numRx_s;

/**
 *  Number of messages to send per second.
 */
@property unsigned int msgRate_s;

/**
 *  Number of messages to send.
 */
@property unsigned int msgNum_s;

/**
 *  Topic to publish to.
 */
@property char *publishTopic_ps;

/**
 *  Logs rx and tx stats for the given session. The session statistics are
 *  cleared before returning.
 *
 *  @param session_p Pointer to the session to print stats for.
 */
- (void)printStatsForSession:(solClient_opaqueSession_pt)session_p;

/**
 *  Callback function that handles the case where a timeout occurs while waiting
 *  waiting for all messages to be received. The number of messages expected to
 *  be received is specified in the application parameter
 *  PARAMETER_NUM_MESSAGES.
 *
 *  @param opaqueContext_p A pointer to the context under which the timer was
 *                         started previously.
 *  @param user_p          A pointer to opaque user data.
 */
- (void)waitRxDoneCallbackWithContext:
            (solClient_opaqueContext_pt)opaqueContext_p
                             userData:(void *)user_p;

/**
 *  The corresponding C function for
 *  waitRxDoneCallbackWithContext:userData. See the documentation for that
 *  method for more information.
 *
 *  @param opaqueContext_p A pointer to the context under which the timer was
 *                         started previously.
 *  @param user_p          A pointer to an instance of the callbackBridge_t
 *                         struct.
 */
static void waitRxDoneCallback(solClient_opaqueContext_pt opaqueContext_p,
                               void *user_p);

/**
 *  This function does the publishing (in its own thread) of the requested
 *  number of messages at the requested rate and then computes the elapsed time.
 *
 *  @param session_p The session on which the messages are published.
 *
 *  @return Return code.
 */
- (threadRetType)pubThread:(void *)session_p;

/**
 *  The corresponding C function for pubThread:. See the documentation for that
 *  method for more information.
 *
 *  @param user_p A pointer to an instance of the callbackBridge_t struct.
 *
 *  @return Return code.
 */
threadRetType pubThread(void *user_p);

/**
 *  Sends multiple messages on the specified Session, which is more efficient
 *  than multiple calls to solClient_session_send().
 *
 *  The purpose of this example is to demonstrate an efficient way for
 *  applications that have several messages to send in one shot. In that case,
 *  bundling N (defined as GROUP_SIZE in in this example) messages into one send
 *  through the solClient_session_sendMultipleMsg(..) will result in one
 *  vectored socket write.
 *
 *  In this case, it is recommended that the Session property
 *  SOLCLIENT_SESSION_PROP_TCP_NODELAY be enabled, since multiple messages are
 *  sent at once onto the underlying TCP connection, and so there is no need to
 *  have the operating system carry out the TCP delay algorithm to cause fuller
 *  packets.
 *
 *  @param session_p The session on which the messages are published.
 *
 *  @return Return code.
 */
- (threadRetType)pubThreadSendMultiple:(void *)session_p;

/**
 *  The corresponding C function for pubThreadSendMultiple:. See the
 *  documentation for that method for more information.
 *
 *  @param user_p A pointer to an instance of the callbackBridge_t struct.
 *
 *  @return Return code.
 */
threadRetType pubThreadSendMultiple(void *user_p);

@end
