/** @example ex/ios/examples/SdtPubSubMsgDepExample.m
 */

/*
 *  Copyright 2009-2025 Solace Corporation. All rights reserved.
 */

#import "SdtPubSubMsgDepExample.h"

@implementation SdtPubSubMsgDepExample

- (id)initWithExampleInterface:(ExampleInterface *)exampleInterface {
    self = [super initWithExampleInterface:exampleInterface];

    // Set example name and description
    self.name = @"SdtPubSubMsgDep";
    self.description =
        @"Demonstrates how to publish SDT map messages to a topic "
        @"(dependent on a Solace message)";

    return self;
}

- (void)run {
    [super run];

    solClient_returnCode_t rc = SOLCLIENT_OK;

    // Context variables
    solClient_opaqueContext_pt context_p;
    solClient_context_createFuncInfo_t contextFuncInfo =
        SOLCLIENT_CONTEXT_CREATEFUNC_INITIALIZER;

    // Session variables
    solClient_opaqueSession_pt session_p;

    // Message variables
    solClient_opaqueMsg_pt msg_p = NULL;
    solClient_destination_t destination;
    int msgsSent = 0;
    solClient_opaqueContainer_pt streamContainer = NULL;
    solClient_opaqueContainer_pt userPropContainer = NULL;
    char messageField[32];

    // Initialize the API; this must be done prior to first usage
    if ((rc = solClient_initialize(SOLCLIENT_LOG_DEFAULT_FILTER, NULL)) !=
        SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_initialize()"];
        goto notInitialized;
    }

    // Set up logging level and log example and API information
    [self setLoggingLevel];
    solClient_log(
        SOLCLIENT_LOG_NOTICE,
        "SdtPubSubMsgDepExample.m (Copyright 2009-2025 Solace Corporation. All rights reserved.)\n");
    [self logCCSMPVersion];

    // Create a context, and specify that the context thread should be created
    // automatically instead of having the application create its own
    // context thread.
    solClient_log(SOLCLIENT_LOG_INFO, "Creating solClient context");
    if ((rc = solClient_context_create(
             SOLCLIENT_CONTEXT_PROPS_DEFAULT_WITH_CREATE_THREAD, &context_p,
             &contextFuncInfo, sizeof(contextFuncInfo))) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_context_create()"];
        goto cleanup;
    }

    /**
     *  Session creation and connection
     */
    solClient_log(SOLCLIENT_LOG_INFO, "Creating solClient session.");

    if ((rc = [self
             createAndConnectSessionWithContext:context_p
                                        session:&session_p
                                messageCallback:messageReceivePrintMsgCallback
                                  eventCallback:eventCallback]) !=
        SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:
                  rc errorString:"createAndConnectSessionWithContext:session:"
                                 "messageCallback:eventCallback:userData:"];
        goto cleanup;
    }

    /**
     *  Subscribe to topic
     */
    if ((rc = solClient_session_topicSubscribeExt(
             session_p, SOLCLIENT_SUBSCRIBE_FLAGS_WAITFORCONFIRM,
             COMMON_MY_SAMPLE_TOPIC)) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_session_topicSubscribe()"];
        goto sessionConnected;
    }

    /**
     *  Message publication
     */
    solClient_log(SOLCLIENT_LOG_INFO, "Publishing messages.\n");

    // Allocate memory for the message to be sent.
    if ((rc = solClient_msg_alloc(&msg_p)) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc errorString:"solClient_msg_alloc()"];
        goto sessionConnected;
    }

    // Set the message delivery mode
    if ((rc = solClient_msg_setDeliveryMode(
             msg_p, SOLCLIENT_DELIVERY_MODE_DIRECT)) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_msg_setDeliveryMode()"];
        goto freeMessage;
    }

    // Set the destination
    destination.destType = SOLCLIENT_TOPIC_DESTINATION;
    destination.dest = COMMON_MY_SAMPLE_TOPIC;
    if ((rc = solClient_msg_setDestination(
             msg_p, &destination, sizeof(destination))) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_msg_setDestination()"];
        goto freeMessage;
    }

    // Create the binary attachment Stream.
    if ((rc = solClient_msg_createBinaryAttachmentStream(
             msg_p, &streamContainer, 1024)) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:
                                "solClient_msg_createBinaryAttachmentStream()"];
        goto freeMessage;
    }

    // Add PI to the Stream.
    if ((rc = solClient_container_addDouble(streamContainer, 3.141592654,
                                            NULL)) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_container_addDouble()"];
        goto freeMessage;
    }

    // Add a String to the Stream.
    if ((rc = solClient_container_addString(streamContainer, "message",
                                            NULL)) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_container_addString()"];
        goto freeMessage;
    }

    // Create a User Property Map; only a MAP may be created in User Properties.
    if ((rc = solClient_msg_createUserPropertyMap(msg_p, &userPropContainer,
                                                  1024)) != SOLCLIENT_OK) {
        [self
            handleErrorWithReturnCode:rc
                          errorString:"solClient_msg_createUserPropertyMap()"];
        goto freeMessage;
    }

    // Add the largest known Mersenne Exponent to the Map.
    if ((rc = solClient_container_addInt32(userPropContainer, 43112609,
                                           "mersenne")) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_container_addInt32()"];
        goto freeMessage;
    }

    for (msgsSent = 0;
         msgsSent < 10 &&
             !self.requestCancel;
         ++msgsSent) {

        if ((rc = solClient_container_deleteField(userPropContainer,
                                                  "message")) != SOLCLIENT_OK) {
            [self
                handleErrorWithReturnCode:rc
                              errorString:"solClient_container_deleteField()"];
            goto freeMessage;
        }

        snprintf(messageField, 32, "message%d", (msgsSent + 1));

        // Populate the Map with a String.
        if ((rc = solClient_container_addString(userPropContainer, messageField,
                                                "message")) != SOLCLIENT_OK) {
            [self handleErrorWithReturnCode:rc
                                errorString:"solClient_container_addString()"];
            goto freeMessage;
        }

        // Send the message
        if ((rc = solClient_session_sendMsg(session_p, msg_p)) !=
            SOLCLIENT_OK) {
            [self handleErrorWithReturnCode:rc
                                errorString:"solClient_session_sendMsg()"];
            goto freeMessage;
        }
    }

freeMessage:
    if ((rc = solClient_msg_free(&msg_p)) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc errorString:"solClient_msg_free()"];
        goto sessionConnected;
    }

    // Wait one second after sending messages. This also gives time for the
    // final message to be received.
    [NSThread sleepForTimeInterval:1];

    /**
     *  Unsubscribe from topic
     */

    if ((rc = solClient_session_topicUnsubscribeExt(
             session_p, SOLCLIENT_SUBSCRIBE_FLAGS_WAITFORCONFIRM,
             COMMON_MY_SAMPLE_TOPIC)) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_session_topicSubscribe()"];
        goto sessionConnected;
    }

/**
 *  Cleanup
 */
sessionConnected:
    // Disconnect the session
    if ((rc = solClient_session_disconnect(session_p)) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_session_disconnect()"];
    }

cleanup:
    // Cleanup solclient
    if ((rc = solClient_cleanup()) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc errorString:"solClient_cleanup()"];
    }

notInitialized:
    [self cleanup];
}

@end
