/** @example ex/ios/examples/SubscribeOnBehalfOfClientExample.m
 */

/*
 *  Copyright 2010-2025 Solace Corporation. All rights reserved.
 */

#import "SubscribeOnBehalfOfClientExample.h"

@implementation SubscribeOnBehalfOfClientExample

- (id)initWithExampleInterface:(ExampleInterface *)exampleInterface {
    self = [super initWithExampleInterface:exampleInterface];

    // Set example name and description
    self.name = @"SubscribeOnBehalfOfClient";
    self.description =
        @"Demonstrates how to subscribe on behalf of another client";

    return self;
}

- (void)run {
    [super run];

    solClient_returnCode_t rc = SOLCLIENT_OK;

    solClient_opaqueContext_pt context_p;
    solClient_context_createFuncInfo_t contextFuncInfo =
        SOLCLIENT_CONTEXT_CREATEFUNC_INITIALIZER;

    solClient_opaqueSession_pt sessionMgr_p;
    solClient_opaqueSession_pt sessionClient_p;

    const char *topic_str = "sample/topic/pasta";
    static const char subscriptionManager[] = "Subscription Manager";
    static const char subscriptionClient[] = "Subscription Client";
    char clientName_a[SOLCLIENT_SESSION_PROP_MAX_CLIENT_NAME_LEN + 1];
    const char *props[40];
    int propIndex;

    // Initialize the API; this must be done prior to first usage
    if ((rc = solClient_initialize(SOLCLIENT_LOG_DEFAULT_FILTER, NULL)) !=
        SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_initialize()"];
        goto notInitialized;
    }

    // Set up logging level and log example and API information
    [self setLoggingLevel];
    solClient_log(
        SOLCLIENT_LOG_NOTICE,
        "SubscribeOnBehalfOfClientExample.m (Copyright 2010-2025 Solace Corporation. All rights reserved. "
        "Systems, Inc. All rights reserved.)\n");
    [self logCCSMPVersion];

    // Create a context, and specify that the context thread be created
    // automatically instead of having the application create its own
    // context thread.
    solClient_log(SOLCLIENT_LOG_INFO, "Creating solClient context");
    if ((rc = solClient_context_create(
             SOLCLIENT_CONTEXT_PROPS_DEFAULT_WITH_CREATE_THREAD, &context_p,
             &contextFuncInfo, sizeof(contextFuncInfo))) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_context_create()"];
        goto cleanup;
    }

    /**
     *  Session creation and connection
     */
    solClient_log(SOLCLIENT_LOG_INFO, "Creating solClient sessions.");

    if ((rc = [self
             createAndConnectSessionWithContext:context_p
                                        session:&sessionMgr_p
                                messageCallback:messageReceivePrintMsgCallback
                                  eventCallback:eventCallback]) !=
        SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:
                  rc errorString:"createAndConnectSessionWithContext:session:"
                                 "messageCallback:eventCallback:userData:"];
        goto cleanup;
    }

    if ((rc = [self
             createAndConnectSessionWithContext:context_p
                                        session:&sessionClient_p
                                messageCallback:messageReceivePrintMsgCallback
                                  eventCallback:eventCallback]) !=
        SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:
                  rc errorString:"createAndConnectSessionWithContext:session:"
                                 "messageCallback:eventCallback:userData:"];
        goto cleanup;
    }

    solClient_log(SOLCLIENT_LOG_INFO, "Checking for capability:  ...");
    if (!solClient_session_isCapable(
            sessionMgr_p, SOLCLIENT_SESSION_CAPABILITY_SUBSCRIPTION_MANAGER)) {
        solClient_log(SOLCLIENT_LOG_ERROR,
                      "Subscription Manager Not Supported. Exiting");
        goto sessionConnected;
    } else {
        solClient_log(SOLCLIENT_LOG_INFO, "OK");
    }

    /**
     *  Get client name for the second session
     */
    if (solClient_session_getProperty(
            sessionClient_p, SOLCLIENT_SESSION_PROP_CLIENT_NAME, clientName_a,
            sizeof(clientName_a)) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_session_getProperty()"];
        goto sessionConnected;
    }

    /**
     *  Subscribe through the session
     */

    solClient_log(SOLCLIENT_LOG_INFO,
                  "Adding subscription %s in %s on behalf of client %s",
                  topic_str, subscriptionManager, subscriptionClient);
    propIndex = 0;
    props[propIndex++] = SOLCLIENT_ENDPOINT_PROP_ID;
    props[propIndex++] = SOLCLIENT_ENDPOINT_PROP_CLIENT_NAME;
    props[propIndex++] = SOLCLIENT_ENDPOINT_PROP_NAME;
    props[propIndex++] = clientName_a;
    props[propIndex++] = NULL;
    if ((rc = solClient_session_endpointTopicSubscribe(
             (char **)props, sessionMgr_p,
             SOLCLIENT_SUBSCRIBE_FLAGS_WAITFORCONFIRM, topic_str, 0)) !=
        SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:
                  rc errorString:"solClient_session_endpointTopicSubscribe()"];
        goto sessionConnected;
    }

    // Send a message on the manager session and see it received on the client
    // only
    if ([self publishMessageWithSession:sessionMgr_p
                                  topic:(char *)topic_str
                           deliveryMode:SOLCLIENT_DELIVERY_MODE_DIRECT] !=
        SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"common_publishDirectMessage()"];
        goto sessionConnected;
    }

    // Wait for half a second

    solClient_log(SOLCLIENT_LOG_NOTICE, "Sent.\n");
    [NSThread sleepForTimeInterval:0.5];
    solClient_log(SOLCLIENT_LOG_NOTICE, "Done.\n");

/**
 *  Cleanup
 */
sessionConnected:
    // Disconnect the session
    if ((rc = solClient_session_disconnect(sessionMgr_p)) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_session_disconnect()"];
    }
    if ((rc = solClient_session_disconnect(sessionClient_p)) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_session_disconnect()"];
    }

cleanup:
    // Cleanup solclient
    if ((rc = solClient_cleanup()) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc errorString:"solClient_cleanup()"];
    }

notInitialized:
    [self cleanup];
}

@end
