/** @example ex/ios/examples/SyncCacheRequestExample.m
 */

/*
 *  Copyright 2009-2025 Solace Corporation. All rights reserved.
 */

#import "SyncCacheRequestExample.h"

@implementation SyncCacheRequestExample

- (id)initWithExampleInterface:(ExampleInterface *)exampleInterface {
    self = [super initWithExampleInterface:exampleInterface];

    // Set example name and description
    self.name = @"SyncCacheRequest";
    self.description = @"Demonstrates how to send a synchronous cache request";

    // Set example parameters
    [self.parameters addParameter:PARAMETER_CACHE_NAME];
    [self.parameters addParameter:PARAMETER_DESTINATION_TOPIC];

    return self;
}

- (void)run {
    [super run];

    solClient_returnCode_t rc = SOLCLIENT_OK;

    // Context.
    solClient_opaqueContext_pt context_p;
    solClient_context_createFuncInfo_t contextFuncInfo =
        SOLCLIENT_CONTEXT_CREATEFUNC_INITIALIZER;

    // Session.
    solClient_opaqueSession_pt session_p;

    // Cache Session.
    solClient_opaqueCacheSession_pt cacheSession_p;
    const char *cacheProps[20];
    int propIndex = 0;
    solClient_cacheRequestFlags_t cacheFlags;
    solClient_uint64_t cacheRequestId = 1;

    // Initialize the API; this must be done prior to first usage
    if ((rc = solClient_initialize(SOLCLIENT_LOG_DEFAULT_FILTER, NULL)) !=
        SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_initialize()"];
        goto notInitialized;
    }

    // Set up logging level and log example and API information
    [self setLoggingLevel];
    solClient_log(
        SOLCLIENT_LOG_NOTICE,
        "SyncCacheRequestExample.m (Copyright 2009-2025 Solace Corporation. All rights reserved.)\n");
    [self logCCSMPVersion];

    // Create a context, and specify that the context thread be created
    // automatically instead of having the application create its own
    // context thread.
    solClient_log(SOLCLIENT_LOG_INFO, "Creating solClient context");
    if ((rc = solClient_context_create(
             SOLCLIENT_CONTEXT_PROPS_DEFAULT_WITH_CREATE_THREAD, &context_p,
             &contextFuncInfo, sizeof(contextFuncInfo))) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_context_create()"];
        goto cleanup;
    }

    /**
     *  Session creation and connection
     */
    solClient_log(SOLCLIENT_LOG_INFO, "Creating solClient sessions.");

    if ((rc = [self
             createAndConnectSessionWithContext:context_p
                                        session:&session_p
                                messageCallback:messageReceivePrintMsgCallback
                                  eventCallback:eventCallback]) !=
        SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:
                  rc errorString:"createAndConnectSessionWithContext:session:"
                                 "messageCallback:eventCallback:userData:"];
        goto cleanup;
    }

    /**
     *  Publish a message
     */

    // Message is published to ensure one is cached
    if ((rc = [self
             publishMessageWithSession:
                 session_p topic:(char *)
                           [[self.parameters
                                parameterWithId:PARAMETER_DESTINATION_TOPIC]
                                   .value
                               cStringUsingEncoding:NSASCIIStringEncoding]
                          deliveryMode:SOLCLIENT_DELIVERY_MODE_DIRECT]) !=
        SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"common_publishMessage()"];
        goto sessionConnected;
    }

    /**
     *  Create a cache session within the connected session
     */

    propIndex = 0;

    cacheProps[propIndex++] = SOLCLIENT_CACHESESSION_PROP_CACHE_NAME;
    cacheProps[propIndex++] =
        [[self.parameters parameterWithId:PARAMETER_CACHE_NAME].value
            cStringUsingEncoding:NSASCIIStringEncoding];

    cacheProps[propIndex] = NULL;

    if ((rc = solClient_session_createCacheSession(
             (const char *const *)cacheProps, session_p, &cacheSession_p)) !=
        SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_session_createCacheSession"];
        goto sessionConnected;
    }

    /**
     *  Cache the request (synchronous)
     */

    solClient_log(SOLCLIENT_LOG_NOTICE, "Sending cache request.\n\n");

    /**
     *  SOLCLIENT_CACHEREQUEST_FLAGS_LIVEDATA_QUEUE:
     *  The cache request completes when the cache response is returned. Live
     *  data matching the cache request Topic is queued until the cache request
     *  completes. The queued live data is delivered to the application before
     *  the cache response message.
     *
     *  SOLCLIENT_CACHEREQUEST_FLAGS_NOWAIT_REPLY:
     *  By default, solClient_cacheSession_sendCacheRequest blocks until the
     *  cache request completes and then returns the cache request status. If
     *  the cache request flag SOLCLIENT_CACHEREQUEST_FLAGS_NOWAIT_REPLY is
     *  set, solClient_cacheSession_sendCacheRequest returns immediately with
     *  the status SOLCLIENT_IN_PROGRESS, and the cache request status is
     *  returned through a callback when it completes.
     *
     *  For more information on other cacheFlags, refer to solCache.h.
     */
    cacheFlags = SOLCLIENT_CACHEREQUEST_FLAGS_LIVEDATA_QUEUE;

    if ((rc = solClient_cacheSession_sendCacheRequest(
             cacheSession_p,
             [[self.parameters parameterWithId:PARAMETER_DESTINATION_TOPIC]
                     .value cStringUsingEncoding:NSASCIIStringEncoding],
             cacheRequestId, NULL, NULL, cacheFlags, 0)) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:
                  rc errorString:"solClient_cacheSession_sendCacheRequest()"];
        goto cleanupCache;
    }

    solClient_log(SOLCLIENT_LOG_NOTICE, "Cache request sent.\n\n");

/**
 *  Cleanup
 */

cleanupCache:
    if ((rc = solClient_cacheSession_destroy(&cacheSession_p)) !=
        SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_cacheSession_destroy()"];
    }

    solClient_log(SOLCLIENT_LOG_NOTICE, "Exiting.\n");

sessionConnected:
    // Disconnect the session
    if ((rc = solClient_session_disconnect(session_p)) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc
                            errorString:"solClient_session_disconnect()"];
    }

cleanup:
    // Cleanup solclient
    if ((rc = solClient_cleanup()) != SOLCLIENT_OK) {
        [self handleErrorWithReturnCode:rc errorString:"solClient_cleanup()"];
    }

notInitialized:
    [self cleanup];
}

@end
